/*
 *                 [[ Frozen Motion ]]
 *
 * Copyright (c) 2004 Glenn Sanson.
 *
 * This code is distributed under the GNU General Public License 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Artwork:
 *    Amaury Amblard-Ladurantie <amaury at linuxfr.org>
 *      (the bubbles)
 *
 * Design & Programming:
 *    Glenn Sanson <glenn.sanson at free.fr>
 *      (whole Java sourcecode, including JIGA classes 
 *             http://glenn.sanson.free.fr/jiga/)
 *
 *          [[ http://glenn.sanson.free.fr/fmotion/home.html ]]
 */

import net.library.jiga.*;

import java.awt.*;

public class BubbleSprite extends Sprite {
	
	final static double SPEED_REDUCTOR = 0.98;
	final static double MAX_SPEED = 5.;
	final static double MIN_COLLISION_SPEED = 0.1;
		
	public final static int TYPE_CURSOR = 0;
	public final static int TYPE_NORMAL = 1;
	public final static int TYPE_MINI = 2;
	public final static int TYPE_COLLISION = 3;
	public final static int TYPE_REMOVED = 4;
	
	final static int NO_COLLISION = -1;
	final static int COLLISION_UP_LEFT = 0;
	final static int COLLISION_UP_RIGHT = 1;
	final static int COLLISION_DOWN_LEFT = 2;
	final static int COLLISION_DOWN_RIGHT = 3;
	
	Image[] bubbleImage;
	SpeedVector movement;
	Game game;
	
	int currentState;
	int cornerCollision;
	
	boolean horizontalExtend;
	boolean verticalExtend;
	
	boolean isMoving;
	
	// Real position
	double realX, realY;
	Rectangle realPosition;
	
	// Points
	Point upLeft, downRight, center;
	
	// Type
	int bubbleType;
	
	// LifeTime
	int time;
	int blinkTime, currentBlink;
	
	public BubbleSprite(Rectangle rect, Image[] image, int type, Game game) {
			
		super(rect);
		
		this.game = game;
		this.bubbleType = type;
		
		this.realX = rect.x;
		this.realY = rect.y;
		this.realPosition = new Rectangle(rect);
				
		this.bubbleImage = image;
		
		this.movement = new SpeedVector();
		isMoving = false;
		
		if (rect.x > game.getGameApplet().getWidth() - rect.width) {
			horizontalExtend = true;
		}

		if (rect.y > game.getGameApplet().getHeight() - rect.height) {
			verticalExtend = true;
		}	
			
		if (type == this.TYPE_MINI) {
			blinkTime = game.getParameters()[Level.PARAM_BLINK_PERIOD_SMALL];
		}
		else { // TYPE_NORMAL
			blinkTime = game.getParameters()[Level.PARAM_BLINK_PERIOD_NORMAL];
		}
		
		cornerCollision = this.NO_COLLISION;
		upLeft = this.wallPosition(realX, realY);
		downRight = this.wallPosition(realX+realPosition.width, realY+realPosition.height);		
	}
	
	public int getType() {
		return bubbleType;
	}
	
	// Return true if bubble is stopped
	public boolean reduceSpeed() {
		movement.resizeVector(SPEED_REDUCTOR);

		return !isMoving;
	}
		
	public boolean isMoving() {
		return isMoving;
	}
	
	// returns true if collision sound is needed
	public boolean moveBubble() {
		this.realX += movement.getX();
		this.realY += movement.getY();
		
		checkBorderPosition();
		
		synchronizePosition();
		
		upLeft = this.wallPosition(realX, realY);
		center = this.wallPosition(realX+realPosition.width/2, realY+realPosition.height/2);
		downRight = this.wallPosition(realX+realPosition.width, realY+realPosition.height);
		
		return checkWallCollision() && (movement.getX()>MIN_COLLISION_SPEED || movement.getY()>MIN_COLLISION_SPEED);
	}
	
	public void partialMove(SpeedVector s, double ratio) {
		this.realX += ratio * s.getX();
		this.realY += ratio * s.getY();
		
		checkBorderPosition();
		
		synchronizePosition();

		upLeft = this.wallPosition(realX, realY);
		center = this.wallPosition(realX+realPosition.width/2, realY+realPosition.height/2);
		downRight = this.wallPosition(realX+realPosition.width, realY+realPosition.height);
				
		checkWallCollision();
	}
	
	void localMove(double posX, double posY) {
		this.realX = posX;
		this.realY = posY;
		
		checkBorderPosition();
		
		synchronizePosition();

		upLeft = this.wallPosition(realX, realY);
		center = this.wallPosition(realX+realPosition.width/2, realY+realPosition.height/2);
		downRight = this.wallPosition(realX+realPosition.width, realY+realPosition.height);		
	}
	
	public double getRealX() {
		return realX;
	}

	public double getRealY() {
		return realY;
	}
		
	public void checkBorderPosition() {
		if (horizontalExtend) {
			if (realX > game.getGameApplet().getWidth()) {
				realX -= game.getGameApplet().getWidth();
				
				this.changeSpriteArea(new Rectangle((int)realX, this.getSpriteArea().y, realPosition.width, this.getSpriteArea().height));
				horizontalExtend = false;
				
				synchronizePosition();
			}
			else if (realX < game.getGameApplet().getWidth() - realPosition.width) {
				this.changeSpriteArea(new Rectangle((int)realX, this.getSpriteArea().y, realPosition.width, this.getSpriteArea().height));
				horizontalExtend = false;				
			}
		}
		else {
			if (realX < 0.) {
				realX += game.getGameApplet().getWidth();
				
				this.changeSpriteArea(new Rectangle(0, this.getSpriteArea().y, game.getGameApplet().getWidth(), this.getSpriteArea().height));
				horizontalExtend = true;
				
				synchronizePosition();
			}
			else if (realX > (game.getGameApplet().getWidth() - realPosition.width)) {
				this.changeSpriteArea(new Rectangle(0, this.getSpriteArea().y, game.getGameApplet().getWidth(), this.getSpriteArea().height));
				horizontalExtend = true;				
			}
		}
		
		if (verticalExtend) {
			if (realY > game.getGameApplet().getHeight()) {
				realY -= game.getGameApplet().getHeight();
				
				this.changeSpriteArea(new Rectangle(this.getSpriteArea().x, (int)realY, this.getSpriteArea().width, realPosition.height));
				verticalExtend = false;
				
				synchronizePosition();
			}
			else if (realY < game.getGameApplet().getHeight() - realPosition.height) {
				this.changeSpriteArea(new Rectangle(this.getSpriteArea().x, (int)realY, this.getSpriteArea().width, realPosition.height));
				verticalExtend = false;				
			}
		}
		else {
			if (realY < 0.) {
				realY += game.getGameApplet().getHeight();
				
				this.changeSpriteArea(new Rectangle(this.getSpriteArea().x, 0, this.getSpriteArea().width, game.getGameApplet().getHeight()));
				verticalExtend = true;
				
				synchronizePosition();
			}
			else if (realY > (game.getGameApplet().getHeight() - realPosition.height)) {
				this.changeSpriteArea(new Rectangle(this.getSpriteArea().x, 0, this.getSpriteArea().width, game.getGameApplet().getHeight()));
				verticalExtend = true;				
			}
		}		
		
		if (horizontalExtend && verticalExtend) {
			this.forceRefresh();
		}
	}
	
	public void synchronizePosition() {
		// Real Movement
		realPosition.setLocation((int)realX, (int)realY);
		
		// Clip area
		if (horizontalExtend) {
			if (!verticalExtend) {
				this.absoluteMove(new Point(0, realPosition.y));
			}
		}
		else {
			if (verticalExtend) {
				this.absoluteMove(new Point(realPosition.x, 0));
			}
			else {
				this.absoluteMove(new Point(realPosition.x, realPosition.y));
			}
		}
	}
	
	public Rectangle getRealPosition() {
		return realPosition;
	}
	
	public boolean isHBorder() {
		return horizontalExtend;
	}
	
	public boolean isVBorder() {
		return verticalExtend;
	}
		
	public boolean repulse(BubbleSprite bubble) {
		
		// Check border
		Rectangle checkRect = new Rectangle(bubble.getRealPosition());
		if (horizontalExtend && checkRect.x < realPosition.width) {
			checkRect.translate(game.getGameApplet().getWidth(), 0);
		}
		else if (bubble.isHBorder() && realPosition.x < checkRect.width) {
			checkRect.translate(-game.getGameApplet().getWidth(), 0);
		}
		
		if (verticalExtend && checkRect.y < realPosition.height) {
			checkRect.translate(0, game.getGameApplet().getHeight());
		}
		else if (bubble.isVBorder() && realPosition.y < checkRect.height) {
			checkRect.translate(0, -game.getGameApplet().getHeight());
		}
		
		// First approx (rectangular areas)
		if (!realPosition.intersects(checkRect)) {
			return false;
		}
		
		// Check for exact collision (sphere radii)
		int r1 = realPosition.width >> 1;
		int r2 = checkRect.width >> 1;
		int sq_radii = (r1+r2)*(r1+r2);
		
		double posX1 = this.getRealX() + r1;
		double posY1 = this.getRealY() + r1;
		double posX2 = checkRect.x + r2;
		double posY2 = checkRect.y + r2;
		
		double dx = posX1 - posX2;
		double dy = posY1 - posY2;
		double sq_dist = dx * dx + dy * dy;
				
		if (sq_dist > sq_radii) {
			return false;
		}
				
		double normalize = 1. / Math.sqrt(dx * dx + dy * dy);
		SpeedVector escape = new SpeedVector(dx, dy);
		
		this.partialMove(escape, normalize);
		bubble.partialMove(escape, -normalize);
		
		return true;
	}
	
	public boolean checkCollision(BubbleSprite bubble) {		
		// Check border
		Rectangle checkRect = new Rectangle(bubble.getRealPosition());
		if (horizontalExtend && checkRect.x < realPosition.width) {
			checkRect.translate(game.getGameApplet().getWidth(), 0);
		}
		else if (bubble.isHBorder() && realPosition.x < checkRect.width) {
			checkRect.translate(-game.getGameApplet().getWidth(), 0);
		}
		
		if (verticalExtend && checkRect.y < realPosition.height) {
			checkRect.translate(0, game.getGameApplet().getHeight());
		}
		else if (bubble.isVBorder() && realPosition.y < checkRect.height) {
			checkRect.translate(0, -game.getGameApplet().getHeight());
		}
		
		// First approx (rectangular areas)
		if (!realPosition.intersects(checkRect)) {
			return false;
		}
				
		// Check for exact collision (sphere radii)
		int r1 = realPosition.width >> 1;
		int r2 = checkRect.width >> 1;
		int sq_radii = (r1+r2)*(r1+r2);
		
		double posX1 = this.getRealX() + r1;
		double posY1 = this.getRealY() + r1;
		double posX2 = checkRect.x + r2;
		double posY2 = checkRect.y + r2;
		
		double dx = posX1 - posX2;
		double dy = posY1 - posY2;
		double sq_dist = dx * dx + dy * dy;
				
		if (sq_dist > sq_radii) {
			return false;
		}
				
		// Collision occured

		// Check appropriate action (maybe no need to process ;)
		if (bubble.getImage() == this.bubbleImage[0]) {
			if (this.getType() == TYPE_NORMAL && bubble.getType() == TYPE_NORMAL) {
				this.collision();
				bubble.collision();
				
				game.eliminateSprite(this);
				game.eliminateSprite(bubble);
				
				return true;
			}
		}
		else {
			if (this.getType() == TYPE_NORMAL && bubble.getType() == TYPE_NORMAL) {
				Rectangle position = new Rectangle((int)((this.getRealX() + checkRect.x)/2.)+8, (int)((this.getRealY() + checkRect.y)/2.)+8, 16, 16);
				Image[] img = game.getMini(this, bubble);
				game.addStaticSprite(new BubbleSprite(position, img, TYPE_MINI, game));
				
			} else if (this.getType() == TYPE_MINI && bubble.getType() == TYPE_CURSOR ) {
				
				game.eliminateSprite(this);
				return true;
			} else if (this.getType() == TYPE_CURSOR && bubble.getType() == TYPE_MINI ) {
				
				game.eliminateSprite(bubble);	
				return true;
			}
		}
		
		SpeedVector other = bubble.getSpeedVector();

		// check when collision occured
		
		// Angle [0 -> line between centers]
		double alpha = Math.PI/2;
		if (dx != 0.) {
			alpha = Math.atan(dy/dx);
		}		
		
		// Restore initial positions
		this.partialMove(this.getSpeedVector(), -1.0);
		bubble.partialMove(other, -1.0);
		
		// Change speed vectors		
		double cos_alpha = Math.cos(alpha);
		double sin_alpha = Math.sin(alpha);
		double cos_c_alpha = cos_alpha * cos_alpha;
		double sin_c_alpha = sin_alpha * sin_alpha;
		double cos_sin_alpha = cos_alpha * sin_alpha;
						
		SpeedVector newCurrent = new SpeedVector(other.x*cos_c_alpha+movement.x*sin_c_alpha+(other.y-movement.y)*cos_sin_alpha,
												 movement.y*cos_c_alpha+other.y*sin_c_alpha+(other.x-movement.x)*cos_sin_alpha);
		SpeedVector newOther = new SpeedVector(movement.x*cos_c_alpha+other.x*sin_c_alpha+(movement.y-other.y)*cos_sin_alpha,
											   other.y*cos_c_alpha+movement.y*sin_c_alpha+(movement.x-other.x)*cos_sin_alpha);
			
		this.setSpeedVector(newCurrent);
		bubble.setSpeedVector(newOther);
		
		// Calculate new position
		this.moveBubble();
		bubble.moveBubble();
		
		repulse(bubble);
		
		return true;
	}
	
	public boolean checkWallCollision() {
						
		boolean collision = false;
		
		if (upLeft.x == downRight.x) {
			if (upLeft.y == downRight.y) {
				// no control needed
			} 
			else {
				if (center.y == upLeft.y) {
					if (game.getWallValue(downRight.x, downRight.y) != 0) {
						
						if (downRight.y != 0) {
							this.localMove(realX, downRight.y*40.-realPosition.height);
						}
						else {
							this.localMove(realX, game.getGameApplet().getHeight()-realPosition.height);
						}
						
						if (movement.y > 0.) {
							this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
						}
						
						collision = true;
					} 
				}
				else {
					if (game.getWallValue(upLeft.x, upLeft.y) != 0) {
						this.localMove(realX, downRight.y*40.);
						
						if (movement.y < 0.) {
							this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
						}
						
						collision = true;
					}					
				}
			}
		}
		else {
			if (upLeft.y == downRight.y) {
				if (center.x == upLeft.x) {
					if (game.getWallValue(downRight.x, downRight.y) != 0) {

						if (downRight.x != 0) {
							this.localMove(downRight.x*40.-realPosition.width, realY);
						} 
						else { 
							this.localMove(game.getGameApplet().getWidth()-realPosition.width, realY);
						}
						
						if (movement.x > 0.) {
							this.setSpeedVector(new SpeedVector(-movement.x, movement.y));
						}
						
						collision = true;
					} 
				}
				else {
					if (game.getWallValue(upLeft.x, upLeft.y) != 0) {
						this.localMove(downRight.x*40., realY);
						
						if (movement.x < 0.) {
							this.setSpeedVector(new SpeedVector(-movement.x, movement.y));
						}
						
						collision = true;
					}
				}
			} 
			else {
				// bubble over 4 Blocks
				if (center.y == upLeft.y) {
					if (center.x == upLeft.x) {
						if (game.getWallValue(downRight.x, upLeft.y) != 0) {
							
							if (downRight.x != 0) {
								this.localMove(downRight.x*40.-realPosition.width, realY);
							} 
							else { 
								this.localMove(game.getGameApplet().getWidth()-realPosition.width, realY);
							}
						
							if (movement.x > 0.) {
								this.setSpeedVector(new SpeedVector(-movement.x, movement.y));
							}
							
							collision = true;
							
							if (game.getWallValue(upLeft.x, downRight.y) != 0) {
								this.localMove(realX, downRight.y*40.-realPosition.height);
									
								if (movement.y > 0.) {
									this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
								}
							}
						}
						else if (game.getWallValue(upLeft.x, downRight.y) != 0) {

							if (downRight.y != 0) {
								this.localMove(realX, downRight.y*40.-realPosition.height);
							}
							else {
								this.localMove(realX, game.getGameApplet().getHeight()-realPosition.height);
							}
									
							if (movement.y > 0.) {
								this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
							}
							
							collision = true;
						}
						else if (game.getWallValue(downRight.x, downRight.y) != 0) {
							
							// Check if collision really occured
							if (Math.pow(realX+realPosition.width/2 - downRight.x*40 - 2, 2.) + Math.pow(realY+realPosition.height/2 - downRight.y*40 - 2, 2.) < (realPosition.width/2)*(realPosition.width/2)) {								
							
								if (cornerCollision != this.COLLISION_DOWN_RIGHT) {
									// Change speedvector direction
									this.setSpeedVector(new SpeedVector(-this.getSpeedVector().getY() / this.SPEED_REDUCTOR, -this.getSpeedVector().getX() / this.SPEED_REDUCTOR));
									
									cornerCollision = this.COLLISION_DOWN_RIGHT;
									return true;
								} 
								
								return false;
							}
						}
					} 
					else { // center.x == downRight.x
						if (game.getWallValue(upLeft.x, upLeft.y) != 0) {
							this.localMove(downRight.x*40., realY);
							
							if (movement.x < 0.) {
								this.setSpeedVector(new SpeedVector(-movement.x, movement.y));
							}
							
							collision = true;

							if (game.getWallValue(downRight.x, downRight.y) != 0) {
								this.localMove(realX, downRight.y*40.-realPosition.height);
									
								if (movement.y > 0.) {
									this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
								}
							}
						}
						else if (game.getWallValue(downRight.x, downRight.y) != 0) {

							if (downRight.y != 0) {
								this.localMove(realX, downRight.y*40.-realPosition.height);
							}
							else {
								this.localMove(realX, game.getGameApplet().getHeight()-realPosition.height);
							}
									
							if (movement.y > 0.) {
								this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
							}
							
							collision = true;
						}
						else if (game.getWallValue(upLeft.x, downRight.y) != 0) {
							
							if (Math.pow(realX+realPosition.width/2 - downRight.x*40 + 2, 2.) + Math.pow(realY+realPosition.height/2 - downRight.y*40 - 2, 2.) < (realPosition.width/2)*(realPosition.width/2)) {								
							
								if (cornerCollision != this.COLLISION_DOWN_LEFT) {
									// Change speedvector direction
									this.setSpeedVector(new SpeedVector(this.getSpeedVector().getY() / this.SPEED_REDUCTOR, this.getSpeedVector().getX() / this.SPEED_REDUCTOR));
									
									cornerCollision = this.COLLISION_DOWN_LEFT;
									return true;
								}
								
								return false;
							}
						}						
					}
				}
				else { // center.y ==downRight.y
					if (center.x == upLeft.x) {
						if (game.getWallValue(downRight.x, downRight.y) != 0) {

							if (downRight.x != 0) {
								this.localMove(downRight.x*40.-realPosition.width, realY);
							} 
							else { 
								this.localMove(game.getGameApplet().getWidth()-realPosition.width, realY);
							}

							if (movement.x > 0.) {
								this.setSpeedVector(new SpeedVector(-movement.x, movement.y));
							}
							
							collision = true;
							
							if (game.getWallValue(upLeft.x, upLeft.y) != 0) {
								this.localMove(realX, downRight.y*40.);
									
								if (movement.y < 0.) {
									this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
								}
							}
						}
						else if (game.getWallValue(upLeft.x, upLeft.y) != 0) {
							this.localMove(realX, downRight.y*40.);
								
							if (movement.y < 0.) {
								this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
							}
							
							collision = true;
						}
						else if (game.getWallValue(downRight.x, upLeft.y) != 0) {
							
							if (Math.pow(realX+realPosition.width/2 - downRight.x*40 - 2, 2.) + Math.pow(realY+realPosition.height/2 - downRight.y*40 + 2, 2.) < (realPosition.width/2)*(realPosition.width/2)) {
								
								if (cornerCollision != this.COLLISION_UP_RIGHT) {
									// Change speedvector direction
									this.setSpeedVector(new SpeedVector(this.getSpeedVector().getY() / this.SPEED_REDUCTOR, this.getSpeedVector().getX() / this.SPEED_REDUCTOR));
									
									cornerCollision = this.COLLISION_UP_RIGHT;
									return true;
								}
								
								return false;
							}
						}
					} 
					else { // center.x == downRight.x
						if (game.getWallValue(upLeft.x, downRight.y) != 0) {
							this.localMove(downRight.x*40., realY);
							
							if (movement.x < 0.) {
								this.setSpeedVector(new SpeedVector(-movement.x, movement.y));
							}
							
							collision = true;

							if (game.getWallValue(downRight.x, upLeft.y) != 0) {
								this.localMove(realX, downRight.y*40.);
									
								if (movement.y < 0.) {
									this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
								}
							}
						}
						else if (game.getWallValue(downRight.x, upLeft.y) != 0) {
							this.localMove(realX, downRight.y*40.);
									
							if (movement.y < 0.) {
								this.setSpeedVector(new SpeedVector(movement.x, -movement.y));
							}
							
							collision = true;
						}
						else if (game.getWallValue(upLeft.x, upLeft.y) != 0) {
							
							// Check if collision really occured
							if (Math.pow(realX+realPosition.width/2 - downRight.x*40 + 2, 2.) + Math.pow(realY+realPosition.height/2 - downRight.y*40 + 2, 2.) < (realPosition.width/2)*(realPosition.width/2)) {
								if (cornerCollision != this.COLLISION_UP_LEFT) {
									// Change speedvector direction
									this.setSpeedVector(new SpeedVector(-this.getSpeedVector().getY() / this.SPEED_REDUCTOR, -this.getSpeedVector().getX() / this.SPEED_REDUCTOR));
									
									cornerCollision = this.COLLISION_UP_LEFT;
									return true;
								}
								
								return false;
							}
						}						
					}
				}
			}	
		}
		
		cornerCollision = this.NO_COLLISION;
		return collision;
	}
	
	Point wallPosition(double x, double y) {
		if (x >= game.getGameApplet().getWidth()) {
			x -= game.getGameApplet().getWidth();
		}
		else if (x < 0.) {
			x += game.getGameApplet().getWidth();
		}

		if (y >= game.getGameApplet().getHeight()) {
			y -= game.getGameApplet().getHeight();
		}
		else if (y < 0.) {
			y += game.getGameApplet().getHeight();
		}
		
		return new Point((int)Math.floor(x/40.), (int)Math.floor(y/40.));
	}	
	
	public SpeedVector getSpeedVector() {
		return movement;
	}
	
	public void setSpeedVector(SpeedVector movement) {
		this.movement = movement;
		
		isMoving = movement.getX() != 0. || movement.getY() != 0.;
	}
		
	public Image getImage() {
		return bubbleImage[0];
	}
	
	public void increaseTime() {
		time++;
	
		if (bubbleType == TYPE_MINI) 
		{
			if (time == game.getParameters()[Level.PARAM_TIME_SMALL]) {
				time = 0;
				currentState = 0;
				blinkTime = game.getParameters()[Level.PARAM_BLINK_PERIOD_NORMAL];
				
				partialMove(new SpeedVector(-8, -8), 1.0);
				bubbleImage = game.getFull(this);
			
				getSpriteArea().setSize(32, 32);
				realPosition.setSize(32, 32);
			
				bubbleType = this.TYPE_NORMAL;
			} else if (time > game.getParameters()[Level.PARAM_TIME_SMALL] - game.getParameters()[Level.PARAM_BLINK_TIME_SMALL]) {
				if (currentBlink == 0) {
					currentBlink = blinkTime;
					currentState = 1 - currentState;
					if (blinkTime > 1 && currentState == 0) {
						blinkTime--;
					}
				}
				
				currentBlink--;
			}
		} else if (bubbleType == TYPE_NORMAL) {
			if (time == game.getParameters()[Level.PARAM_TIME_NORMAL]) {
				game.addExplosion(this.getRealPosition(), game.getMini(bubbleImage[0]));				
				game.eliminateSprite(this);
			} else if (time > game.getParameters()[Level.PARAM_TIME_NORMAL] - game.getParameters()[Level.PARAM_BLINK_TIME_NORMAL]) {
				if (currentBlink == 0) {
					currentBlink = blinkTime;
					currentState = 1 - currentState;
					if (blinkTime > 1 && currentState == 0) {
						blinkTime--;
					}
				}
				
				currentBlink--;
			}
		}
	}
	
	public void remove() {
		bubbleType = this.TYPE_REMOVED;
	}
	
	public void collision() {
		bubbleType = this.TYPE_COLLISION;
	}	
	
	public void paint(Graphics g, GameApplet applet) {
		
		g.setClip(this.getSpriteArea());
		
		g.drawImage(bubbleImage[currentState], realPosition.x, realPosition.y, applet);
		
		if (horizontalExtend) {			
			g.drawImage(bubbleImage[currentState], realPosition.x-applet.getWidth(), realPosition.y, applet);
			
			if (verticalExtend) {
				g.drawImage(bubbleImage[currentState], realPosition.x-applet.getWidth(), realPosition.y-applet.getHeight(), applet);
			}
		}
		
		if (verticalExtend) {
			g.drawImage(bubbleImage[currentState], realPosition.x, realPosition.y-applet.getHeight(), applet);
		}
	}
}