/*
 *                 [[ Frozen Motion ]]
 *
 * Copyright (c) 2004 Glenn Sanson.
 *
 * This code is distributed under the GNU General Public License 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Artwork:
 *    Amaury Amblard-Ladurantie <amaury at linuxfr.org>
 *      (the bubbles)
 *
 * Design & Programming:
 *    Glenn Sanson <glenn.sanson at free.fr>
 *      (whole Java sourcecode, including JIGA classes 
 *             http://glenn.sanson.free.fr/jiga/)
 *
 *          [[ http://glenn.sanson.free.fr/fmotion/home.html ]]
 */

import java.util.Vector;

public class Level {
	
	private final static String BEGIN_TAG = "[";
	private final static String END_TAG = "]";
	private final static String SEPARATOR = ":";	
	
	private final static String DATA_NAME = "N";
	private final static String DATA_AUTHOR = "A";
	private final static String DATA_PARAMS = "P";
	private final static String DATA_WALLS = "W";
	private final static String DATA_BUBBLES = "B";
	private final static String DATA_LINK = "L";
	private final static String DATA_DIFFICULTY = "D";
	
	public final static int PARAM_TOTAL_TIME = 0;
	public final static int PARAM_TIME_INC = 1;
	public final static int PARAM_TIME_DEC = 2;
	public final static int PARAM_TIME_SMALL = 3;
	public final static int PARAM_BLINK_PERIOD_SMALL = 4;
	public final static int PARAM_BLINK_TIME_SMALL = 5;
	public final static int PARAM_TIME_NORMAL = 6;
	public final static int PARAM_BLINK_PERIOD_NORMAL = 7;
	public final static int PARAM_BLINK_TIME_NORMAL = 8;

	final static int[] DEFAULT_PARAMETERS = { 4000, 100, 600, 600, 11, 140, 2400, 23, 560 };
	
	String name;
	String author;
	int[] parameters;
	int difficulty;
	byte[][] walls;
	Vector bubbles;
	Vector links;
	boolean completed;
	
	public Level(String level) {
		
		parameters = DEFAULT_PARAMETERS;
		walls = new byte[16][12];
		bubbles = new Vector();
		links = new Vector();
		author = new String();
		
		int begin = level.indexOf(BEGIN_TAG);
		
		while (begin != -1) {
			int end = level.indexOf(END_TAG, begin);
			
			if (end == -1) {
				System.err.println("Bad argument : "+level.substring(begin));
			} else {
				checkArgument(level.substring(begin+BEGIN_TAG.length(), end));
				begin = level.indexOf(BEGIN_TAG, end);
			}			
		}
	}
	
	public boolean isValid() {
		return ((name != null) && (bubbles.size()>=2));
	}
	
	void checkArgument(String text) {
		Vector checked = getData(text);
		
		if (checked.size() < 2) {
			System.err.println("Incomplete argument : ["+text+"]");
			return;
		}
		
		String type = (String)checked.elementAt(0);
		
		if (type.equals(this.DATA_NAME)) {
			if (checked.size() > 2) {
				System.err.println("Too many arguments for LEVEL NAME : ["+text+"]");
				return;
			}	
			
			name = (String)checked.elementAt(1);
		}		
		else if (type.equals(this.DATA_AUTHOR)) {
			if (checked.size() > 2) {
				System.err.println("Too many arguments for LEVEL AUTHOR : ["+text+"]");
				return;
			}

			author = (String)checked.elementAt(1);
		} 
		else if (type.equals(this.DATA_PARAMS)) {
			if (checked.size() != 6) {
				System.err.println("Bad number of arguments for LEVEL PARAMETERS : ["+text+"]");
				return;
			}
			
			int[] newParameters = new int[9];
			
			for (int i=0 ; i<5 ; i++) {
				try {
					newParameters[i] = Integer.parseInt((String)checked.elementAt(i+1));
				}
				catch (Exception e) {
					System.err.println("Invalid parameter for LEVEL PARAMETERS : ["+text+"]");
					return;
				}
			}
			
			// Add Blink
			newParameters[6] = newParameters[4];
			int[] blink = calculateBlink(newParameters[3]);
			newParameters[4] = blink[0];
			newParameters[5] = blink[1];
			blink = calculateBlink(newParameters[6]);
			newParameters[7] = blink[0];
			newParameters[8] = blink[1];			
			
			parameters = newParameters;
		}
		else if (type.equals(this.DATA_DIFFICULTY)) {
			if (checked.size() > 2) {
				System.err.println("Too many arguments for DIFFICULTY : ["+text+"]");
				return;
			}
			try {
				difficulty = Integer.parseInt((String)checked.elementAt(1));
			} 
			catch(Exception e) {
				System.err.println("Invalid parameter for DIFFICULTY : ["+text+"]");
			}
		} 
		else if (type.equals(this.DATA_WALLS)) {
			if (checked.size() > 2) {
				System.err.println("Too many arguments for WALLS MAP DEFINITON : ["+text+"]");
				return;
			}
			if (((String)checked.elementAt(1)).length() != 192) {
				System.err.println("Wrong size of walls map : ["+text+"]");
				return;			
			}
			
			byte[][] newWalls = new byte[16][12];
			
			int position = 0;
			for (int j=0 ; j<12 ; j++) {
				for (int i=0 ; i<16 ; i++) {
					try {
						newWalls[i][j] = Byte.parseByte(((String)checked.elementAt(1)).substring(position, position+1));
					}
					catch(Exception e) {
						System.err.println("Invalid parameter for WALLS MAP DEFINITION : ["+text+"]");
						return;						
					}
					
					position++;
				}
			}
			
			walls = newWalls;
		}
		else if (type.equals(this.DATA_BUBBLES)) {
			if (checked.size() != 4) {
				System.err.println("Bad number of arguments for BUBBLE DEFINITION : ["+text+"]");
				return;
			}
			
			int[] newBubble = new int[3];
			
			try {
				newBubble[0] = Integer.parseInt((String)checked.elementAt(1)); // Color
				newBubble[1] = Integer.parseInt((String)checked.elementAt(2)) - 16; // X Coord
				newBubble[2] = Integer.parseInt((String)checked.elementAt(3)) - 16; // Y Coord
			}
			catch (Exception e) {
				System.err.println("Invalid parameter for BUBBLE DEFINITION : ["+text+"]");
				return;						
			}
			
			bubbles.addElement(newBubble);
		}
		else if (type.equals(this.DATA_LINK)) {
			if (checked.size() != 3) {
				System.err.println("Bad number of arguments for LINK DEFINITION : ["+text+"]");
				return;
			}
			
			int[] newLink = new int[2];
			
			try {
				newLink[0] = Integer.parseInt((String)checked.elementAt(1));
				newLink[1] = Integer.parseInt((String)checked.elementAt(2));
			}
			catch (Exception e) {
				System.err.println("Invalid parameter for LINK DEFINITION : ["+text+"]");
				return;						
			}
			
			if (Math.max(newLink[0], newLink[1]) > bubbles.size()) {
				System.err.println("Link Error (Bubble not defined) : ["+text+"]");
				return;
			}
			
			links.addElement(newLink);
		}
		else {
			System.err.println("Unknown type of parameter : ["+text+"]");
			return;
		}
	}
	
	Vector getData(String text) {
		Vector output = new Vector();
		
		int begin = 0;
		int end =  text.indexOf(SEPARATOR);
		
		while (end != -1) {
			
			output.addElement(text.substring(begin, end));

			begin = end+SEPARATOR.length();
			end = text.indexOf(SEPARATOR, begin);	
			
			if (end == -1) {
				output.addElement(text.substring(begin));	
			}			
		}
		
		return output;
	}

	int[] calculateBlink(int time) {
		int output[] = new int[2];
	
		time >>= 2;
		// Blink period
		output[0] = 1;
		// Blink total duration
		output[1] = 10;
		
		while (output[1]+2*(output[0]+1) < time) {
			output[0]++;
			output[1] += 2*output[0];
		}
		
		return output;
	}
	
	public void setFinished() {
		completed = true;
	}
	
	public byte[][] getWalls() {
		return walls;
	}
	
	public Vector getBubbles() {
		return bubbles;
	}
	
	public Vector getLinks() {
		return links;
	}	
	
	public int[] getParameters() {
		return parameters;
	}
	
	public int getDifficulty() {
		return difficulty;
	}
	
	public String getName() {
		return name;
	}
	
	public String getAuthor() {
		return author;
	}
	
	public boolean isCompleted() {
		return completed;
	}
	
	public void reset() {
		completed = false;
	}
}